;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

(defclass kimmo-automaton-set (filename-symbol-mixin ;what was the NIL here?
                               kimmo-string-mixin
                               trace-flag-mixin
                               property-list-mixin)
  ;; Inherited IV name-symbol will eval to the kimmo-automaton-set.
  ;; a list \a, \b, ..., S, V, ....
  ((char-list :accessor char-list 
              :initarg :char-list)
   ;; highest index used by alphabet chars and subset names
   (char-interp-limit :accessor char-interp-limit
                      :initarg :char-interp-limit)
   ;; maps index to one of:
   ;;     (ALPHABET char),
   ;;     (NULL char),
   ;;     (ANY ... chars ...),
   ;;     (SUBSET ... chars ...).
   ;; CDR of interp indicates the chars
   ;;     that can match the char.  other
   ;;     flavors depend on this.
   (char-interp-array :accessor char-interp-array
                      :initarg :char-interp-array)
   ;; a list of (LexChar.SurfChar) pairs
   (feasible-pairs :accessor feasible-pairs
                   :initarg :feasible-pairs)
   (automaton-flavor :initform 'kimmo-automaton
                     :initarg :automaton-flavor
                     :accessor automaton-flavor)
   (recognizer-flavor :initform 'kimmo-recognizer
                      :initarg :recognizer-flavor
                      :accessor recognizer-flavor)
   (generator-flavor :initform 'kimmo-generator
                     :initarg :generator-flavor
                     :accessor generator-flavor)
   ;; list of automaton objects
   (automaton-list :accessor automaton-list
                   :initform nil
                   :initarg :automaton-list)
   ;; not necessarily in automaton-list
   (merged-automaton :accessor merged-automaton
                     :initarg :merged-automaton) ;probably computed...
   (merged-automaton-name :initform '|Merged Automaton|
                          :initarg :merged-automaton-name
                          :accessor merged-automaton-name)
   (normal-recognizer :accessor normal-recognizer
                      :initform nil)
   (merged-recognizer :accessor merged-recognizer
                      :initform nil)
   (normal-generator  :accessor normal-generator
                      :initform nil)
   (merged-generator  :accessor merged-generator
                      :initform nil)
   ;; NORMAL or MERGED
   (mode :initform 'normal
         :accessor mode)
   ;; either the normal one or the merged one.
   (recognizer :accessor recognizer)
   ;; either the normal one or the merged one.
   (generator :accessor generator)
   ;; e.g., ENGLISH-DICT
   (dictname :accessor dictname
             :initarg dictname
             :initform nil)
   ;; (open close)
   (balanced-comment-delimiters :accessor balanced-comment-delimiters))
  ;;(:init-options suppress-initial-read)
  )

(defmethod default-dictionary ((self kimmo-automaton-set) name)
  (etypecase name
    (null
     ;; use the default
     (unless (and (slot-boundp self 'dictname) (dictname self))
       (error "No dictionary name has been specified."))
     (get-defined-value (dictname self))
     (symbol-value (dictname self)))
    (symbol
     (setf (dictname self) name)
     (get-defined-value name)
     (symbol-value name))
    (kimmo-dictionary
     (setq name (name-symbol name))
     (setf (dictname self) name)
     (get-defined-value name)
     (symbol-value name))))

;;;
;;; Slot accessors and the like
;;;

(defmethod automata ((self kimmo-automaton-set))
  ;; evidently wants to refer to it by this name, too.
  (automaton-list self))

(defmethod retrieve-automaton ((self kimmo-automaton-set) name)
  ;; find an automaton with a particular name
  (find name (automaton-list self) :key #'name :test #'equal))

;;; Before you choke, note will one day be initialize-instance.
(defmethod pcl::initialize :after ((self kimmo-automaton-set) plist)
  (get-defined-value '*kimmo-reader*)
  (unless (getf plist :suppress-initial-read)
    (read-file self)))

(defmacro kimmo-index (sym)
  ;; say what?  this lookslike an even worse kludge...
  `(get ,sym 'kimmo-index))

;;; The "char" here is obviously really a symbol.  Convert to chars & hashtable?
(defmethod assign-char-index ((self kimmo-automaton-set) sym)
  ;; For efficiency, these indices are global rather than per-language.
  (check-type sym symbol)
  (or (kimmo-index sym)
      ;; Yow.  Talk about re-inventing wheels.  Out-of-round ones, at that...
      (let ((index (or (get 'kimmo-index 'next-index) 0)))
	(setf (get 'kimmo-index 'next-index) (1+ index))
	(setf (kimmo-index sym) index)
	index)))

;;;
;;; Storing character interpretations
;;;

(defmethod show-char-interps ((self kimmo-automaton-set))
  (loop for x in (char-list self)
        for interp = (char-interp self x)
        do (format t "~&Character ~S: ~S~%" x interp)))

(defmethod char-interp ((self kimmo-automaton-set) char &optional class)
  (let* ((index (kimmo-index char))
	 (interp (and index
		      (<= index (char-interp-limit self))
                      (aref (char-interp-array self) index))))
    (when (and (eq class 'must-be-defined)
	       (null interp))
      ;; that's what I like to see: a little more constructive paranoia! -- sgr
      (error  "Character ~S is not defined." char))
    (cond ((null class) interp)
	  ((eq class 'must-be-defined) interp)
	  ((eq (car interp) class) interp)
	  ((and (listp class) (member (car interp) class)) interp)
	  (t nil))))

(defmethod alphabet ((self kimmo-automaton-set))
  ;; previous implementors have opined that this is slow.
  (loop for x in (char-list self)
        when (char-interp self x 'alphabet)
          collect x))

;;; Should this be a SETF method?  (Probably; one problem at a time, son.)
(defmethod set-char-interp ((self kimmo-automaton-set) char interp)
  (let* ((index (or (kimmo-index char)
                    (error "There is no kimmo index for ~S." char)))
	 (a (char-interp-array self))
         (old (aref a index)))
    (assert (<= index (char-interp-limit self)) (index))
    (cond ((equal old interp) nil)
	  ((null interp)
	   (error "Assigning a null interpretation to ~S." char))
	  ((null old)
	   (setf (char-list self)
		 (nconc (char-list self) (cons char nil)))
           (setf (aref a index) interp))
	  (t (error "Character ~S can't be interpreted as both ~S and ~S."
                    char old interp)))))

(defmethod match-pair ((self kimmo-automaton-set) pair pattern)
  ;; PAIR contains real characters; PATTERN may not
  (and (member (car pair) 
               (cdr (or (char-interp self (car pattern))
                        (error "~S has no interpretation in ~S."
                               (car pattern) self))))
       (member (cdr pair)
               (cdr (or (char-interp self (cdr pattern))
                        (error "~S has no interpretation in ~S."
                               (cdr pattern) self))))))

(defmethod explode-string-verify ((self kimmo-automaton-set) string)
  (let ((char-list* (explode-string self string)))
    (loop for x in char-list*
       unless (char-interp self x '(alphabet null))
       do (error "String ~S contains character ~S, which is illegal~
                ~&because it's not null and not in the alphabet."
                 string x))
    char-list*))

;;;
;;; Feasible pairs
;;;

(defmethod recompute-feasibles ((self kimmo-automaton-set))
  (setf (feasible-pairs self) nil)
  ;; guessing at the meaning of mapc-message...
  (loop for automaton in (automaton-list self)
        doing (send-feasible-pairs automaton self 'add-feasible-pair)))

(defmethod validate-actual-pair ((self kimmo-automaton-set) pair)
  ;; disallow subsets, etc.
  (unless (and (char-interp self (car pair) 'alphabet)
	       (char-interp self (cdr pair) '(alphabet null)))
    (error "~S is an invalid actual pair." pair)))

(defmethod add-feasible-pair ((self kimmo-automaton-set) pair)
  (unless (member pair (feasible-pairs self) :test #'equal)
    (validate-actual-pair self pair)
    (push pair (feasible-pairs self))))

;;;
;;; Selecting the machine type
;;;

(defmethod forget-machines ((self kimmo-automaton-set))
  ;; hit yourself over the head with this here baseball bat, sonny...
  (setf (normal-recognizer self) nil)
  (setf (normal-generator  self) nil)
  (setf (merged-automaton  self) nil)
  (setf (merged-recognizer self) nil)
  (setf (merged-generator  self) nil)
  (setf (mode self) 'normal))

(defmethod set-up-machines ((self kimmo-automaton-set) &optional set-up-mode)
  (when set-up-mode
    (check-type set-up-mode (member NORMAL MERGED))
    (setf (mode self) set-up-mode))
  (ecase (mode self)
    (NORMAL
      (unless (and (normal-recognizer self)
		   (normal-generator  self))
	(build-normal-machines self))
      (setf (recognizer self) (normal-recognizer self))
      (setf (generator self)  (normal-generator self)))
    (MERGED
      (unless (and (merged-recognizer self)
		   (merged-generator  self))
	(build-merged-machines self))
      (setf (recognizer self) (merged-recognizer self))
      (setf (generator self) (merged-generator self)))))

(defmethod state-vectors ((self kimmo-automaton-set) pair &optional warn-p)
  (loop for aut in (automaton-list self)
        for vec = (state-vector aut pair)
        when (null vec) 
        do (when warn-p
	    (format *error-output*
                    "~&Warning: Feasible pair ~S is dead in automaton ~A.~%"
                    pair (name aut)))
           (return nil)
        collect vec))

(defmethod feasible-pair-data ((self kimmo-automaton-set))
  (loop for pair in (or (feasible-pairs self)
                        (error "There are no feasible-pairs in ~S." self))
     collect (cons pair (state-vectors self pair t))))

;;;
;;; Building the normal machines.  (Recognizer & generator.)
;;;

(defmethod build-normal-machines ((self kimmo-automaton-set))
  (format t "~&Building normal machines...~%")
  (let* ((initial (mapcar #'initial-state (automaton-list self)))
	 (fvecs (mapcar #'finality-vector (automaton-list self)))
	 (rec (make-instance (recognizer-flavor self)
                             :automaton-set self
                             :automaton-list (automaton-list self)
                             :finality-vectors fvecs
                             :initial-config initial))
	 (gen (make-instance (generator-flavor self)
                             :automaton-set self
                             :automaton-list (automaton-list self)
                             :finality-vectors fvecs
                             :initial-config initial))
	 (pair-data (feasible-pair-data self)))
    (absorb-machine-data rec pair-data)
    (absorb-machine-data gen pair-data)
    (setf (normal-recognizer self) rec)
    (setf (normal-generator self) gen)))

;;;
;;; Building the merged machine
;;;

(defmethod build-merged-automaton ((self kimmo-automaton-set))
  ;; is this special only because of Maclisp lossage?
  (declare (special *n-stored-zeros*)) ;proclaim if this is used elsewhere...
  (unless (normal-recognizer self)
    (build-normal-machines self))
  (format t "~&Building merged automaton...~%")
  (let* ((rec (normal-recognizer self))
	 (initial (initial-config rec))
	 (dead (mapcar #'(lambda (ignore) 
                           (declare (ignore ignore))
                           0) 
                       initial))
	 ;; The NCONC element of each column contains a list
         ;; of (merged) state numbers.
	 (columns
	    (loop for pair in (feasible-pairs self)
                  for nconc = (make-instance 'nconc-object)
                  do (add nconc 0)
                  collect (list pair (state-vectors self pair t) nconc)))
	 (finalities (make-instance 'nconc-object)) 
	 (config-queue (make-instance 'closure-queue ;*** closure-queue
                                      :storage-size 100.))
	 (dead-index (prog1 (add config-queue dead)
			    (get-next config-queue)))
	 (n-entries 0)
	 (n-zeros 0)
	 (*n-stored-zeros* 0)
	 (vector-length-cell (list nil))
	 (vector-maker (make-instance 'basic-hashing-interner ;***
			       :storage-size 100.
			       :create-function
                               #'(lambda (data index)
                                   (declare (ignore index))
                                   (let ((a (make-array (car vector-length-cell)
                                                        :initial-contents data)))
                                     (loop for x in data
                                           when (= x 0)
                                           do (incf *n-stored-zeros*))
                                     a))))
	 n fvec bigaut)
    (declare (fixnum n-entries n-zeros))
    (add finalities nil)
    (unless (equal dead-index 0)
      (error "Nonzero state number ~S assigned to dead configuration."
             dead-index))
    (add config-queue initial)
    ;; fill out the "table" stored in COLUMNS
    (loop as config = (get-next config-queue)
          while config
          do (add finalities (config-final rec config))
             ;; add an entry for CONFIG to each column
             (loop for (pair vectors nconc) in columns
                   as nconfig = (advance-config rec config vectors)
                   as nindex = (if (null nconfig) 
                                 0
                                 (add config-queue nconfig))
                   do (add nconc nindex)
                      (incf n-entries)
                      (when (= nindex 0)
                        (incf n-zeros))))
    ;; now convert the columns of the table into vectors, after noting
    ;; the vector length
    (setq n (1+ (length finalities)))
    (setf (car vector-length-cell) n)
    (setq fvec (make-array n :initial-contents (contents finalities)))
    (loop for (pair vectors nconc) in columns
          as transitions = (contents nconc)
          as vec = (intern vector-maker transitions)
          collect (list pair vec) into pair-data
          finally 
            ;; make the automaton
            (setq bigaut
                  (make-instance (automaton-flavor self)
                                 :name (merged-automaton-name self)
                                 :n-rows (- n 2)
                                 :n-columns (length (feasible-pairs self))
                                 :finality-vector fvec
                                 :automaton-set self
                                 :neutral-data pair-data)))
    (setf (merged-automaton self) bigaut)
    (format t "~&Merged automaton table statistics:~
               ~&    ~D rows, ~D columns.~
               ~&    ~D virtual entries in table.~
               ~&    ~D nonzero virtual entries.~
               ~&    ~D distinct transition vectors.~
               ~&    ~D stored entries in table.~
               ~&    ~D nonzero stored entries.~%"
            (- n 2)	; subtract 1 for 1+ above, then subtract 1 for dummy
                        ; initial entry
            (length (feasible-pairs self))
            n-entries
            (- n-entries n-zeros)
            (next-index vector-maker)
            (* (next-index vector-maker) n)
            *n-stored-zeros*)))

(defmethod build-merged-machines ((self kimmo-automaton-set))
  (unless (merged-automaton self)
    (build-merged-automaton self))
  (format t "~&Building merged machines...~%")
  (let* ((bigaut (or (merged-automaton self)
                     (error "There is no merged-automaton in ~S." self)))
	 (bigrec
          (make-instance (recognizer-flavor self)
                         :automaton-set self
                         :automaton-list (list bigaut)
                         :finality-vectors (list (finality-vector bigaut))
                         :initial-config (list (initial-state bigaut))))
	 (biggen
          (make-instance (generator-flavor self)
                         :automaton-set self
                         :automaton-list (list bigaut)
                         :finality-vectors (list (finality-vector bigaut))
                         :initial-config (list (initial-state bigaut))))
	 (pair-data (neutral-data bigaut)))
    (absorb-machine-data bigrec pair-data)
    (absorb-machine-data biggen pair-data)
    (setf (merged-generator  self) biggen)
    (setf (merged-recognizer self) bigrec)))

;;;
;;; Running the automata
;;;

(defmethod generate ((self kimmo-automaton-set) string
                     &key
                     (tracing (trace-flag self))
                     ((:mode mode*) 'normal))
  (set-up-machines self mode*)
  (generate (generator self) 
            (explode-string-verify self string)
            tracing))

(defmethod recognize ((self kimmo-automaton-set) string 
                      &key
                      (tracing (trace-flag self))
                      ((:dictname dictname*) nil)
                      ((:mode mode*) 'normal))
  (set-up-machines self mode*)
  (recognize (recognizer self)
             (explode-string-verify self string)
             (default-dictionary self dictname*)
             tracing))

(defmethod prompt-for-strings ((self kimmo-automaton-set) format-string)
  ;; allow users the pleasures of their input editors.
  (format t format-string)
  (let ((line (string (read-line))))
    (with-input-from-string (s line)
      (read-until *kimmo-reader* '*CR* s))))

(defmethod generator-loop ((self kimmo-automaton-set) &rest keywords)
  (enable-special-return *kimmo-reader* '*CR*)
  (mapc #'(lambda (char) (enable-alphabetic-char *kimmo-reader* char))
        '(#\. #\: #\;))
  (disable-comments *kimmo-reader*)
  (loop for strings = (prompt-for-strings self "~&Lexical strings: ")
;	(progn (format t "~&Lexical strings: ")
;	       (read-until *kimmo-reader* '*CR*))
        while strings
        do (loop for string in strings
                 for results = (apply #'generate self string keywords)
                 if results
                   do (format t "~&~5T~S -> ~S" (string string) (car results))
                      (when (cdr results)
                        (loop with indent = (+ 5 2 4 (length (string string)))
                              for surf in (cdr results)
                              doing (format t "~&~vT~S" indent surf)))
                 else do (format t "~&~5T~S rejected." (string string))))
  (values))

(defmethod recognizer-loop ((self kimmo-automaton-set) &rest keywords)
  (enable-special-return *kimmo-reader* '*CR*)
  (mapc #'(lambda (char) (enable-alphabetic-char *kimmo-reader* char))
        '(#\. #\: #\;))
  (disable-comments *kimmo-reader*)
  (loop for strings = (prompt-for-strings self "~&Surface strings: ")
;	(progn (format t "~&Surface strings: ")
;	       (read-until *kimmo-reader* '*CR*))
        while strings
        do (loop for string in strings
                 for results = (apply #'recognize self string keywords)
                 if results
                   do (format t "~&~5T~S -> ~S ~S" 
                              (string string) (caar results) (cadar results))
                      (when (cdr results)
                        (loop with indent = (+ 5 2 4 (length (string string)))
                              for (lex props) in (cdr results)
                              doing (format t "~&~vT~S ~S" indent lex props)))
                 else do (format t "~&~5T~S rejected." (string string))))
  (values))

(defmethod generator-demo ((self kimmo-automaton-set) 
                           &optional 
                           (strings (getprop self '%generator-demo) 
                                    strings-supplied-p)
                           &rest keywords)
  (cond ((null strings)
         ;; nothing to do anyway, so figure out why
         (if strings-supplied-p
           (format t "~&You seem to be asking me to generate from no strings???")
           (format t "~&No %GENERATOR-DEMO supplied in .AUT file for ~A." self)))
        (t
         ;; there are strings, so we can try generation
         (loop with max-string-length = (loop for string in strings
                                              maximize (length string))
               ;; of course, if you have a symbolics, formatting-table is better
               initially (setq max-string-length 
                               (1+ (max max-string-length (length "rejected"))))
                         (format t "~&Lexical~vTSurface~
                                    ~&=======~:*~vT======="
                                 max-string-length)
               for string in strings
               for results = (apply #'generate
                                    self string
                                    :allow-other-keys t
                                    keywords)
               do (format t "~&~A~vT " string max-string-length)
                  (if (null results)
                    (princ "rejected")
                    (print-separated-list results nil))
                  (terpri))))
  (values))

(defmethod recognizer-demo ((self kimmo-automaton-set)
                            &optional 
                            (strings (when (and (slot-boundp self 'dictname)
                                                (boundp (dictname self))
                                                (typep (symbol-value (dictname self))
                                                       'kimmo-dictionary))
                                       (getprop (symbol-value (dictname self))
                                                '%recognizer-demo))
                                     strings-supplied-p)
                            dictionary
                            &rest keywords)
  (setq dictionary (default-dictionary self dictionary))
  (cond ((null strings)
         ;; nothing to do anyway, so figure out why
         (if strings-supplied-p
           (format t "~&You seem to be asking me to recognize no strings???")
           (format t "~&No %RECOGNIZER-DEMO supplied in .DIC file for ~A." dictionary)))
        (t
         ;; there are strings, so we can try generation
         (loop with max-string-length = (loop for string in strings
                                              maximize (length string))
               ;; of course, if you have a symbolics, formatting-table is better
               initially (setq max-string-length 
                               (1+ (max max-string-length (length "rejected"))))
               (format t "~&Surface~vTLexical~
                          ~&=======~:*~vT======="
                       max-string-length)
               for string in strings
               for results = (apply #'recognize
                                    self string
                                    :allow-other-keys t
                                    :dictname dictionary
                                    keywords)
               do (format t "~&~A~vT " string max-string-length)
               (if (null results)
                 (princ "rejected")
                 (print-separated-list results nil))
               (terpri))))
  (values))
;(:select-nth-alternative :argument
;                         "rejected~*"
;                         (:use-repeatedly-on-list "~S")
;                         ;;(:use-repeatedly-on-list "~S, ~S")
;                         "results ~N")

;;;
;;; File Input
;;;


(defmethod keyword-parse ((self kimmo-automaton-set) keywords endmarker stream)
  (let ((clauses nil))
    (loop as x = (read-next-thing *kimmo-reader* stream nil)
       do (cond ((eq x endmarker)
		 (return (nreverse clauses)))
		((member x keywords)
		 (push (list x) clauses))
		((null clauses)
		 (file-syntax-error self "The file doesn't begin with a keyword."))
		(t (nconc (car clauses) (list x)))))))

(defmethod process-file-header ((self kimmo-automaton-set) stream)
  (disable-digits *kimmo-reader*)
  (disable-special-return *kimmo-reader*)
  (mapc #'(lambda (char) (enable-alphabetic-char *kimmo-reader* char))
        '(#\. #\: #\;))
  (disable-comments *kimmo-reader*)
  (setf (balanced-comment-delimiters self) nil)
  (setf (accent-codes self) nil)
  (let ((clauses 
	  (keyword-parse self '(ALPHABET NULL ANY SUBSET
                                %COMMENT-DELIMITERS %ACCENTS)
                         'END stream))
	max-index alphabet delim open accents)
    (cond ((eq (caar clauses) '%COMMENT-DELIMITERS)
	   (setq delim (cdr (pop clauses)))
           (setf (first delim)  (coerce (first delim)  'character)
                 (second delim) (coerce (second delim) 'character))
	   (unless (and (= 2 (length delim))
			(not (eq (car delim) (cadr delim))))
	     (file-syntax-error self "Bad `%COMMENT-DELIMITERS open close' declaration."))
	   (setf (balanced-comment-delimiters self) delim)
           (enable-comments *kimmo-reader* (first delim) (second delim))
           (setq open (first delim))
	   (loop for clause in clauses
                 when (member open clause)
                   do (file-syntax-error self "Restriction: Comments can't be used in file header.")))
	  ((assoc '%COMMENT-DELIMITERS clauses)
	   (file-syntax-error self "Restriction: %COMMENT-DELIMITERS must be first if present.")))
    ;; Process the %ACCENTS declaration if any, and remove it.
    (when (setq accents (assoc '%ACCENTS clauses))
      (setf (accent-codes self) 
	    (loop for x in (cdr accents)
                  collect (aref (string x) 0))) ;(car (exploden x))))
      (setq clauses (delete accents clauses :test #'equal)))
    ;; Now all remaining clauses involve alphabet characters.
    (unless (eq (caar clauses) 'ALPHABET)
      (file-syntax-error self "The file doesn't begin with an ALPHABET declaration."))
    (setf (char-list self) nil)
    (setq max-index
      (loop for c in (cons '(NULL \0) clauses)
            maximize (loop for x in (cdr c)
                           maximize (assign-char-index self x))))
    (setf (char-interp-limit self) max-index)
    (setf (char-interp-array self) (make-array (1+ max-index))) 
    (setf (feasible-pairs self) nil)
    (setq alphabet (cdr (car clauses)))
    (loop for x in alphabet
          do (set-char-interp self x `(alphabet ,x)))
    (set-char-interp self '\0 '(null \0))
    (loop for (keyword . args) in (cdr clauses)
          do (ecase keyword
               (ALPHABET
                (file-syntax-error self "Duplicate ALPHABET declaration."))
               (NULL
                (unless (equal args '(\0))
                  (file-syntax-error self "'NULL 0' is the only legal NULL declaration in this implementation.")))
               (ANY
                (loop for x in args
                      do (set-char-interp self x `(ANY \0 ,@alphabet))))
               (SUBSET
                (loop for x in (cdr args)
                      unless (member x alphabet)
                      do (file-syntax-error self "Invalid char ~S in SUBSET declaration." x))
                (set-char-interp self (or (car args)
                                          (error "(Car args) is NIL?"))
                                 (cons 'SUBSET (cdr args))))))))

(defmethod read-file ((self kimmo-automaton-set))
  (with-open-stream (stream (input-stream self))
    (format t "~&Reading ~A...~%" (filename self))
    (process-file-header self stream)
    (setf (automaton-list self) nil)
    (forget-machines self)
    (loop with (x y aut)
       do (enable-special-return *kimmo-reader* '*CR*)
          (mapc #'(lambda (char) (enable-alphabetic-char *kimmo-reader* char))
                '(#\. #\: #\;))
          (setq x (read-next-thing *kimmo-reader* stream nil))
          (etypecase x
            ((member END) (return nil))
            ((member %GENERATOR-DEMO)
             (let ((token (return-token *kimmo-reader*)))
               (disable-special-return *kimmo-reader*)
               (setf (getprop self '%GENERATOR-DEMO)
                     (read-until *kimmo-reader* 'end stream))
               (enable-special-return *kimmo-reader* token)))
	    ((member %RECOGNIZER-DEMO)
             (file-syntax-error self "%RECOGNIZER-DEMO belongs in the .DIC file."))
	    ((member %LISP-PROPERTY)
             ;; this uses the regular lisp reader, not the kimmo reader.
             (setf (getprop self (read stream)) (read stream)))
            ((member *CR*)) ;just skip blank lines
	    (string
             (setq y (intern x)) ;gack. -- sgr
             (when (retrieve-automaton self y)
               (file-syntax-error self	"~S is a duplicate automaton name." x))
             (setq aut (make-instance (automaton-flavor self)
                                      :automaton-set self
                                      :name y))
             (if (eq y (merged-automaton-name self))
               (setf (merged-automaton self) aut)
               (setf (automaton-list self)
                     (nconc (automaton-list self) (list aut))))
             (read-self aut stream))))
    (recompute-feasibles self)))

;;;
;;; File Output
;;;

;;; not implemented yet

;;; Har! -- sgr
