;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

(defclass kimmo-automaton (print-object-mixin)
  ;; internal symbol w/ same spelling as string, e.g., #:|Gemination|
  ((name            :initarg :name
                    :accessor name)
   (n-rows          :accessor n-rows ; number of states
                    :initarg :n-rows)
   (n-columns       :accessor n-columns ; number of columns
                    :initarg :n-columns)
   (finality-vector :accessor finality-vector ;array of T/NIL elements (boolean?)
                    :initarg :finality-vector)
   (automaton-set   :accessor automaton-set ;backpointer to kimmo-automaton-set
                    :initarg :automaton-set)
   ;; list of pairs ((lex . surf) state-vector), where state-vector is an array
   ;; with n-rows + 1 elements.  This is the primary representation.  R-data and
   ;; G-data in the automaton-set are derived from it.
   (neutral-data    :accessor neutral-data
                    :initarg :neutral-data)))

(defmethod state-final ((self kimmo-automaton) state-number)
  (aref (finality-vector self) state-number))

(defmethod send-feasible-pairs ((self kimmo-automaton) recipient msg)
  (loop for (pair vector) in (neutral-data self)
        when (and (eq (car (char-interp (automaton-set self) (car pair)))
                      'alphabet)
                  (member (car (char-interp (automaton-set self) (cdr pair)))
                          '(null alphabet)))
        ;; pair is (alphabet . null) or (alphabet . alphabet)
        do (funcall msg recipient pair)))

(defmethod state-vector ((self kimmo-automaton) pair)
  (loop for (pattern vector) in (neutral-data self)
     when (match-pair (automaton-set self) pair pattern)
       return vector))

(defmethod initial-state ((self kimmo-automaton))
  1)

(defmethod %pair-ordering-kludge ((self kimmo-automaton) pattern+vector)
  (let ((p (car pattern+vector)))
    (cons 
     (+ (* 1000. (length (cdr (char-interp (automaton-set self)
                                           (car p) 'must-be-defined))))
	(length (cdr (char-interp (automaton-set self)
                                  (cdr p) 'must-be-defined))))
     pattern+vector)))

;;; not used any more
(defmethod %pair-more-specific ((self kimmo-automaton) p1 p2)
  (and (not (equal p1 p2))
       (let ((x1 (cdr (char-interp (automaton-set self) 
                                   (car p1) 'must-be-defined)))
	     (x2 (cdr (char-interp (automaton-set self) 
                                   (cdr p1) 'must-be-defined)))
	     (y1 (cdr (char-interp (automaton-set self) 
                                   (car p2) 'must-be-defined)))
	     (y2 (cdr (char-interp (automaton-set self) 
                                   (cdr p2) 'must-be-defined))))
	 (and (or (equal x1 y1)
		  (< (length x1) (length y1)))
	      (or (equal x2 y2)
		  (< (length x2) (length y2)))))))

;;;
;;; File Input
;;;

(defmethod read-self ((self kimmo-automaton) stream)
  ;; how to read an automaton from a file.
  (setf (neutral-data self) nil)
  (enable-special-return *kimmo-reader* '*CR*)
  (enable-digits *kimmo-reader*)
  ;; read rows, columns; this implementation allows * instead of a number.
  (let ((tokens (read-until *kimmo-reader* '*CR* stream))
        lex-row surf-row finalities rows)
    (setf (n-rows self) (or (first tokens) '*))
    (setf (n-columns self) (or (second tokens) '*))
    (disable-digits *kimmo-reader*)
    (mapc #'(lambda (char) (enable-alphabetic-char *kimmo-reader* char))
          '(#\. #\: #\;))
    (setq lex-row (read-until *kimmo-reader* '*CR* stream)
          surf-row (read-until *kimmo-reader* '*CR* stream))
    (unless (= (length lex-row) (length surf-row))
      (file-syntax-error self "Lexical and surface rows have different lengths:~
                               ~&~S~
                               ~&~S."
                         lex-row surf-row))
    (when (eq (n-columns self) '*)
      (setf (n-columns self) (length lex-row)))
    (unless (= (length lex-row) (n-columns self))
      (file-syntax-error self "Declared number of columns is wrong."))
    (unless (> (length lex-row) 0)
      (file-syntax-error self "There appear to be no columns."))
    ;; read the rows
    (enable-digits *kimmo-reader*)
    ;; *** (set-single *kimmo-reader* '(#\. #\: #\;)) 
    (mapc #'(lambda (char) (enable-delimiter-char *kimmo-reader* char))
          '(#\. #\: #\;))
    (loop with state and fin and row
          for i from 1
          doing (setq tokens (read-until *kimmo-reader* '*CR* stream))
          until (null tokens)
          do (setq state (first tokens)
                   fin   (second tokens)
                   row   (cddr tokens))
             (unless (equal state i)
               (file-syntax-error self "Row for state ~D is out of order." state))
             (ecase fin
               (#\: (push t finalities))
               (#\. (push nil finalities)))
             (unless (= (length row) (n-columns self))
               (file-syntax-error self "Row for state ~D is the wrong length." state))
             (push row rows))
    (setq rows (nreverse rows)
          finalities (nreverse finalities))
    ;; check n-rows
    (when (eq (n-rows self) '*)
      (setf (n-rows self) (length rows)))
    (unless (= (n-rows self) (length rows))
      (file-syntax-error self "Declared number of rows is wrong."))
    ;; set the neutral-data and finality-vector slots
    (setf (finality-vector self) 
          (make-array (1+ (n-rows self))
                      :initial-contents (cons nil finalities)))
    (setf (neutral-data self)
          (sort-neutral-data 
           self
           (loop with pair and column
                 for i from 1 to (n-columns self)
                 collect
                 (progn
                   (setq pair (cons (pop lex-row) (pop surf-row)))
                   (setq column (loop for rowt on rows
                                      as x = (pop (car rowt))
                                      unless (and (typep x 'fixnum)
                                                  (>= x 0)
                                                  (<= x (n-rows self)))
                                        do (file-syntax-error self "~D is an invalid next state." x)
                                      collect x))
                   (list pair
                         (make-array (1+ (n-rows self))
                                     :initial-contents (cons -1 column)))))))))

(defmethod sort-neutral-data ((self kimmo-automaton) data)
  ;; don't use SORTCAR and %pair-more-specific for this --- it assumes a
  ;; total ordering and will screw up.  this algorithm still isn't the
  ;; right one but it is better than previous attempts.
  (mapcar #'cdr 
          (sort (mapcar #'(lambda (x) (%pair-ordering-kludge self x)) data)
                #'<
                :key #'car)))

;;;
;;; File Output
;;;

(defmethod write-self ((self kimmo-automaton) stream)
  ;; set up base etc.
  (call *kimmo-reader* #'write-self-internal self stream))

;;; Of course, if you have a Symbolics lispm, you'll want to use
;;; formatting-table instead of a kludge like this! -- sgr
(defmethod write-self-internal ((self kimmo-automaton) stream)
  (let ((widths
	  (cons (+ 2 (flatc (n-rows self)))
		(loop for (pair column) in (neutral-data self)
		   collect
		     (1+ (loop for x in (list* (car pair)
					       (cdr pair)
					       (cdr (coerce column 'list)))
			    maximize (flatc x)))))))
    (format stream "~%\"~A\"    ~D ~D~%"
	    (name self) (n-rows self) (n-columns self))
    (format t "~v@T" (car widths))
    (loop for (pair column) in (neutral-data self)
          for wid in (cdr widths)
          do (format stream "~v@A" wid (car pair)))
    (format t "~%~v@T" (car widths))
    (loop for (pair column) in (neutral-data self)
          for wid in (cdr widths)
          do (format stream "~v@A" wid (cdr pair)))
    (format t "~%")
    (loop for i from 1 to (n-rows self)
       do (format t "~v@A~C" (1- (car widths)) i 
		  (if (state-final self i) #\: #\.))
          (loop for (pair column) in (neutral-data self)
                for wid in (cdr widths)
                do (format stream "~v@A" wid (aref column i)))
	  (format t "~%"))))

;;;
;;; Miscellaneous
;;;

(defmethod print-object-name ((self kimmo-automaton))
  ;; the idiosyncratic part of printing one of these
  (if (slot-boundp self 'name)
    (name self)
    "Unnamed"))

(defmethod file-syntax-error ((self kimmo-automaton) fmt &rest args)
  ;; how to complain about the tast of the file you're chewing on.
  (error "Syntax error while reading automaton ~A:~&~?" (name self) fmt args))
