;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

;;;
;;; MacLisp Nostalgia Department.  You can tell this is a quick port, no?
;;;

;;; Apparently this isn't called anymore???
;(defmethod getl ((plist-or-symbol t) prop-names) 
;  ;; see p. 2-53 of the Blue Moonual.  Sigh.
;  ;; This is generic as a courtesy to property-list-mixin, below.
;  (multiple-value-bind (indicator-found value-found plist-tail)
;                       (get-properties (if (symbolp plist-or-symbol)
;                                         ;; MacLisp-style DWIM
;                                         (symbol-plist plist-or-symbol)
;                                         plist-or-symbol)
;                                       prop-names)
;    (declare (ignore indicator-found value-found))
;    plist-tail))

(defun get-defined-value (name)
  ;; make sure this symbol is bound.  (MacLisp tried to autoload if not...)
  ;; (sigh -- assert is a macro)
  (unless (boundp name)
    (error "~S is unbound." name)))

;;;
;;; For making system functions generic.  (The old definition ends up as a 
;;; method on T.
;;;

(eval-when (eval load compile)
(defun generify-function (fn arglist)
  (let (#+:coral (ccl:*warn-if-redefine-kernel* nil)) ;gag fascist warnings
    (pcl::make-specializable fn :arglist arglist)))
)

;;;
;;; si:property-list-mixin
;;;

(defclass property-list-mixin () 
  ((plist :initform nil
          :accessor plist
          :initarg :plist)))

;;; Yes, I could make this method get instead of getprop (e.g., remprop below).
;;; However, there appear to be either PCL or Coral bugs in generifying
;;; setf methods like (setf get).  (It clobbers the system one, and I don't know
;;; how to generify it so it doesn't.) -- sgr

(defmethod getprop ((self property-list-mixin) prop)
  ;; how to look something up in a property-list instance.
  (getf (plist self) prop))

(defmethod (setf getprop) (val (self property-list-mixin) prop)
  ;; how to set a property's value in a property-list instance.
  (setf (getf (plist self) prop) val))

(eval-when (eval load compile)
(generify-function 'remprop '(symbol property-name))
)
(defmethod remprop ((self property-list-mixin) property-name)
  (remf (plist self) property-name))

;(defmethod getl ((self property-list-mixin) props)
;  ;; give our plist to the default method
;  (getl (plist self) props))

;(defmethod snarf-properties ((self property-list-mixin) props plst)
;  ;; takes a bunch of props from plst and copies them to self's plist.
;  (loop for prop in props
;        for prop-value = (getf plst prop)
;        when prop-value
;          do (setf (getf (plist self) prop) prop-value)))

;;;
;;; Mixin for common case of printing self with a type and a name.
;;; You have to supply a method for print-object-name.  (Too bad CLOS doesn't 
;;; support protocols, i.e., :required-methods option to defclass.)
;;;

(defclass print-object-mixin () ())

(defmethod print-object ((self print-object-mixin) stream)
  ;; print class name and something idiosyncratic
  ;; I suppose this could be done (cryptically) as an :around method.
  (pcl::printing-random-thing (self stream)
    (format stream "~S ~A" 
            (class-name (class-of self))
            (print-object-name self))))

;;;
;;; I/O.  Too bad format's not generic.  If you've got a Symbolics, you're 
;;; better off with format-textual-list.
;;;

(defun print-separated-list (list &optional
                                  (slashify-p t)
                                  (separator ", ")
                                  (stream *standard-output*))
  ;; useful for printing lists of items.  Too bad FORMAT's not generic; if it
  ;; were, you'd define a format directive for this.
  (loop for rest = (rest list) then (rest rest)
        for item in list
        doing (if slashify-p
                (prin1 item stream)
                (princ item stream))
        when rest do (princ separator stream)))

(defun flatc (x)
  ;; for compatibility with yucky old MacLisp
  ;; if there were a real stream protocol, we could do a better job of this.
  (length
   (with-output-to-string (s)
     (princ x s))))

;;;
;;; NCONC-OBJECT is used (evidently) by the automaton-set code.
;;;

(defclass nconc-object ()
  ((head :accessor head)
   (tail :accessor tail)))

(defmethod pcl::initialize ((self nconc-object) plist)
  (declare (ignore plist))
  (reset self))

(defmethod reset ((self nconc-object))
  (let ((cell (list nil)))
    ;; this is a headed list
    (setf (head self) cell)
    (setf (tail self) cell)))

(defmethod add ((self nconc-object) x)
  (let ((tl (or (tail self)
                (error "Tail of ~S is NIL?" self)))
	(cell (list x)))
    (unless (null (cdr tl))
      ;; might have been interrupted out of,
      ;; or might have had an add-to-front operation
      (setq tl (last tl)))
    (setf (cdr tl) cell)
    (setf (tail self) cell)
    x))

(defmethod delq ((self nconc-object) x)
  (setf (cdr (head self)) (delete x (cdr (head self))))
  (setf (tail self) (last (head self))))

(defmethod contents ((self nconc-object))
  (cdr (or (head self) (error "Head of ~S is NIL?" self))))

(defmethod add-to-front ((self nconc-object) x)
  (push x (cdr (or (head self) (error "Head of ~S is NIL?" self)))))

(defmethod nconc-it ((self nconc-object) list2)
  (let ((tl (or (tail self) (error "Tail of ~S is NIL?" self))))
    (nconc tl list2)
    (setq tl (last tl))
    (assert (typep tl 'list) (tl))
    (setf (tail self) tl)
    nil))

(defmethod nconc-length ((self nconc-object))
  (length (cdr (or (head self) (error "Head of ~S is NIL?" self)))))

(defmethod nth-tail ((self nconc-object) n &optional initial-value)
  ;; this always creates the tail
  (assert (>= n 0) (n))
  (or (nthcdr (1+ n) (or (head self) (error "HEAD of ~S is NIL?" self)))
      (loop for i from (length (cdr (head self))) by 1
            while (<= i n)
              collect initial-value into ntail
            finally (nconc-it self ntail)
                    (return (last ntail)))))

(defmethod print-object ((self nconc-object) stream)
  ;; how to print yourself.
  (pcl::printing-random-thing (self stream)
    (format stream "(~D element~:P)" (nconc-length self))))

;;;
;;; Used by things that want to have a trace flag.
;;;

(defclass trace-flag-mixin ()
  ((trace-flag :accessor trace-flag
               :initform nil
               :initarg :trace-flag)))

(defclass tracer-storage-mixin ()
  ((tracer-storage :accessor tracer-storage
                   :initform (make-hash-table))
   (default-tracer-flavor :accessor default-tracer-flavor
     :initarg :default-tracer-flavor
     :initform nil)
   (previous-tracer :accessor previous-tracer
                    :initform nil)))

(defmethod get-tracer ((self tracer-storage-mixin) flavor-name)
  (cond ((null flavor-name) nil)
        ;; pcl::object?? gack. -- sgr
	((typep flavor-name 'pcl::object);(instance-p flavor-name)
	 (setf (previous-tracer self) flavor-name))
        ((not (eq flavor-name 't))
	 (setf (previous-tracer self) 
	       (retrieve-tracer self flavor-name)))
	((null (previous-tracer self))
	 (setf (previous-tracer self) 
	       (retrieve-tracer self (default-tracer-flavor self))))
	(t nil))
  (previous-tracer self))

(defmethod retrieve-tracer ((self tracer-storage-mixin) flavor-name)
  (multiple-value-bind (value found-p found-key)
                       (gethash flavor-name (tracer-storage self))
    (declare (ignore found-key))
    (when (not found-p)
      ;; it's not already in the table, so create it
      (setq value (make-tracer self flavor-name))
      (setf (gethash flavor-name (tracer-storage self)) value))
    value))

(defmethod make-tracer ((self tracer-storage-mixin) flavor-name)
  (make-instance flavor-name))

;;;
;;; I finally acceded to the inevitable and converted this, rather than 
;;; try to figure out each call.
;;;

;;; (:select-nth-alternative <n> ... <alternatives> ...), zero-based
;;;			; <n> is :argument or :n-remaining-arguments
;;;			; the last one is the default
;;; (:if-arguments-remain <if> <else>)
;;; (:if-argument-null <if> <else>)
;;;		; Arg is gobbled in <IF>, available to <ELSE>.
;;; (:if-argument-condition <if> <else>)
;;;		; Arg is not available to either
;;; (:unless-argument-null <stuff>)
;;;		; Arg is available to <STUFF>, gobbled o.w.
;;; (:use-repeatedly-on-list ... <stuff> ...)
;;; (:use-repeatedly-on-arguments ... <stuff> ...)
;;; (:use-on-each-list-argument ... <stuff> ...)
;;; :take-format-string-and-arglist
;;; (:use-on-elements-inserting-spaces <spec>)
;;; (:separated-list <element> <separator>)
;;; (:english-list <element> &optional <comma-space>)
;;; :ignore-argument
;;; :call-as-function			; may print to standard-output
;;; :argument-and-call-function		; ditto
;;; :one-argument-and-call-function	; ditto
;;; :two-arguments-and-call-function	; ditto
;;; (:progn ... <stuff> ...)

(defmacro format-template (&rest specs)
  (let ((str (make-array 100.
                         :element-type 'string-char
                         :fill-pointer 0)))
    (write-format-template specs str)
    (string-append str)))

(defvar *bad-format-template "~S is an invalid FORMAT-TEMPLATE spec.")

(defun write-format-template (specs str)
  (loop for spec in specs
     if (stringp spec) 
       do (write-string-template spec str)
     else if (symbolp spec)
       do (write-special-template spec nil str)
     else if (listp spec) 
       do (write-special-template (car spec) (cdr spec) str)
     else do (error *bad-format-template spec)))

(defun string-append (&rest strs)
  (apply #'concatenate 'string strs))

(defun write-string-template (x str)
  (format str "~A" x))

(defun write-special-template (name args str)
  (let ((f (get name 'write-format-template)))
    (when (null f)
      (error *bad-format-template (cons name args)))
    (apply f str args)))

(defun numeric-argument-string (n)
  (etypecase n
    (null "")
    ((member :argument) "V")
    ((member :n-remaining-arguments) "#")
    (fixnum (format nil "~D" n))))

(defmacro defun-property ((symbol property) arglist &body body)
  ;; how to define functions on the plist
  (let ((name (gensym)))
    `(progn (defun ,name ,arglist ,@body)
            (setf (get ',symbol ',property) #',name)
            ',symbol)))

(defun-property (:select-nth-alternative write-format-template)
  (str n &rest specs)
  (unless (member n '(:argument :n-remaining-arguments))
    (error "~S is an invalid value for N." n))
  (format str "~~~A[" (numeric-argument-string n))
  (loop for spectail on specs
     do (write-format-template (list (car spectail)) str)
        (cond ((cddr spectail)
	       (format str "~~;"))
	      ((cdr spectail)
	       (format str "~~:;"))))
  (format str "~~]"))

(defun-property (:if-arguments-remain write-format-template)
  (str spec1 &optional spec2)
  (write-format-template `((:select-nth-alternative :n-remaining-arguments 
			       ,(or spec2 "") ,spec1))
			 str))

(defun-property (:if-argument-null write-format-template)
  (str spec1 &optional spec2)
  (format str "~~:[")
  (write-format-template (list spec1) str)
  (format str "~~;")
  (when spec2 (write-format-template (list "~:*" spec2) str))
  (format str "~~]"))

(defun-property (:if-argument-condition write-format-template)
  (str spec1 &optional spec2)
  (format str "~~:[")
  (when spec2 (write-format-template (list spec2) str))
  (format str "~~;")
  (write-format-template (list spec1) str)
  (format str "~~]"))

(defun-property (:unless-argument-null write-format-template) (str spec)
  (format str "~~@[")
  (write-format-template (list spec) str)
  (format str "~~]"))

(defun-property (:use-repeatedly-on-list write-format-template)
  (str &rest specs)
  (format str "~~{")
  (write-format-template specs str)
  (format str "~~}"))

(defun-property (:use-repeatedly-on-arguments write-format-template)
  (str &rest specs)
  (format str "~~@{")
  (write-format-template specs str)
  (format str "~~}"))

(defun-property (:use-repeatedly-on-sublists write-format-template)
  (str &rest specs)
  (format str "~~:{")
  (write-format-template specs str)
  (format str "~~}"))

(defun-property (:use-on-each-list-argument write-format-template)
  (str &rest specs)
  (format str "~~:@{")
  (write-format-template specs str)
  (format str "~~}"))

(defun-property (:take-format-string-and-arglist write-format-template) (str)
  (format str "~~1{~~:}"))

(defun-property (:use-on-elements-inserting-spaces write-format-template)
  (str spec)
  (write-format-template
    `((:use-repeatedly-on-list
         ,spec (:if-arguments-remain " ")))
    str))

(defun-property (:separated-list write-format-template) (str element separator)
  (write-format-template
    `((:use-repeatedly-on-list
         ,element (:if-arguments-remain ,separator)))
    str))

(defun-property (:english-list write-format-template) 
       (str element &optional (comma-space ", "))
  (write-format-template
   `((:use-repeatedly-on-list
        ,element (:select-nth-alternative :n-remaining-arguments
		    "" (:progn ,comma-space "and ") ,comma-space)))
   str))

(defun-property (:ignore-argument write-format-template) (str)
  (format str "~~*"))

(defun-property (:call-as-function write-format-template) (str)
  (format str "~~Q"))

(defun-property (:argument-and-call-function write-format-template) (str)
  (format str "~~VQ"))

(defun-property (:one-argument-and-call-function write-format-template) (str)
  (format str "~~VQ"))

(defun-property (:two-arguments-and-call-function write-format-template) (str)
  (format str "~~V,VQ"))

(defun-property (:progn write-format-template) (str &rest specs)
  (write-format-template specs str))

;;;
;;; Yet another instance of "convert in order to avoid thinking about what
;;; this is really trying to do."
;;; 

#-Symbolics
(defmacro desetq (&rest x)
  ;; destructuring setq
  (let ((loop::loop-desetq-temporary nil))
    (let ((setq-form (loop::loop-make-desetq x)))
      (if loop::loop-desetq-temporary
        `((lambda (,loop::loop-desetq-temporary) ,setq-form) nil)
        setq-form))))

#+Symbolics
(defmacro desetq (&rest x)
  `(zl:desetq ,@x))

(defmacro continuation-search (vars &body forms-and-options)
  (macrolet ((append! (values var)
	       `(setq ,var (append ,var ,values)))
	     (caseq? (x &rest clauses)
	       `(case ,x
		  ,@ (loop for (match . actions) in clauses
			collect `(,match ,@ actions 't))
		  (t nil)))
	     (progify (list)
	       `(list* 'prog 'nil
		       (loop for x in ,list
			  unless (null x)
			    collect x)))
		)
    (prog (body-forms
	   finally-forms
	   include-stats print-stats-if
	   incf-setq-form incf-enqueue-form finally-stats-form
	   include-level include-step print-steps-if
	   print-continuations-if print-gensym-continuation
	   print-spread-continuation
	   print-setq-continuation print-enqueue-continuation
	   print-pop-continuation print-start-continuation
	   incf-step-form each-step-form
	   user-vars setq-vars
	   gather-maker-list spread-list spread-form gather-maker setq-maker
	   unsetq-form
	   sys-let-clauses sys-let-fixnum sys-declaration
	   can-setq-test continue-macro-body continue-macro
	   enqueue-form
	   funcall-continue-var sys-let-special)
      ;; Initialize the storage-layout variables.
      (setq user-vars vars)
      (setq gather-maker-list vars)	;; used in CONTINUE macro
      (setq spread-list vars)		;; DESETQ-style
      ;; Assume depth-first search.
      (setq can-setq-test '(not continuation-setq-flag))
      (setq enqueue-form '(push backtracking-state backtracking-queue))
      ;; Next, process the options.
      (loop for x in forms-and-options
	 when (not (listp x))
	   collect x into frm
	 else when (caseq? (car x)
		     (:finally (append! (cdr x) finally-forms))
		     (:print-statistics-if (setq print-stats-if (cadr x)))
		     (:print-statistics (setq print-stats-if 't))
		     (:include-statistics (setq include-stats t))
		     (:include-level (setq include-level t))
		     (:include-step (setq include-step t))
		     (:print-steps (setq print-steps-if t))
		     (:print-steps-if (setq print-steps-if (cadr x)))
		     (:print-continuations (setq print-continuations-if t))
		     (:print-continuations-if (setq print-continuations-if
						    (cadr x)))
		     (:depth-first
		        ;; it's the default
		       )
		     (:breadth-first
		      (setq can-setq-test 
			    '(and (null backtracking-queue)
				  (not continuation-setq-flag)))
		      (append! '((backtracking-queue-tail nil))
			       sys-let-clauses)
		      (setq enqueue-form
			    '(cond ((null backtracking-queue)
				    (setq backtracking-queue
					  (setq backtracking-queue-tail
						(ncons backtracking-state))))
				   (t (setf (cdr backtracking-queue-tail)
					    (ncons
					     backtracking-state))
				      (setq backtracking-queue-tail
					    (cdr backtracking-queue-tail)))))
		      )
		     (:special-continue-funcall
		       (setq funcall-continue-var (cadr x))
		         )
		     )
	   do nil
	 else collect x into frm
	finally (setq body-forms frm))
      ;; Do some delayed option processing.
      (when print-stats-if
	(setq include-stats t)
	(setq finally-stats-form
	  '(format t "~&Setq count: ~D~%Enqueue count: ~D~%"
		     continuation-setq-count continuation-enqueue-count))
	(unless (eq print-stats-if 't)
	  (setq finally-stats-form 
		`(when ,print-stats-if ,finally-stats-form)))
        )
      (when include-stats 
	(append! '((continuation-setq-count 0)
		   (continuation-enqueue-count 0))
		 sys-let-clauses)
	(append! '(continuation-setq-count continuation-enqueue-count)
		 sys-let-fixnum)
	(setq incf-setq-form 
	      '(setq continuation-setq-count (1+ continuation-setq-count)))
	(setq incf-enqueue-form
	      '(setq continuation-enqueue-count
		     (1+ continuation-enqueue-count)))
	 )
      (when print-steps-if 
	(setq include-level t)
	(setq include-step t)
	(setq each-step-form
	  '(if backtracking-flag
	       (format t "~&Step ~D, level ~D, continuing from step ~D.~%"
			 continuation-step
			 continuation-level
			 continuation-source-step)
	       (format t "~&Step ~D, level ~D.~%"
			 continuation-step
			 continuation-level)))
	(unless (eq print-steps-if 't)
	  (setq each-step-form `(when ,print-steps-if ,each-step-form)))
	)
      (when include-level
	(append! '((continuation-level 0)) sys-let-clauses)
	(append! '(continuation-level) sys-let-fixnum)
	(append! '('(1+ continuation-level)) gather-maker-list)
	(append! '(continuation-level) spread-list)
	)
      (when include-step
	(append! '((continuation-step 0)
		   (continuation-source-step -1)) sys-let-clauses)
	(append! '(continuation-step continuation-source-step) sys-let-fixnum)
	(append! '('continuation-step) gather-maker-list)
	(append! '(continuation-source-step) spread-list)
	(setq incf-step-form '(setq continuation-step (1+ continuation-step)))
	)
      ;; Finalize storage layout.
      (setq gather-maker
            `(list 'setq 'backtracking-state
                   (cons 'list (list ,@ gather-maker-list))))
      (setq spread-form
            `(desetq ,spread-list backtracking-state))
      (setq setq-vars 
	    (loop for x in spread-list
	       collect (gensym)))
      (setq setq-maker
	    `(list 'setq 
		   ,@ (loop for value in gather-maker-list
			     as var in setq-vars
			     collect `(quote ,var)
			     collect value)))
      (setq unsetq-form
	    `(setq ,@ (loop for value in setq-vars
			     as var in spread-list
			     collect var
			     collect value)))
      (when funcall-continue-var
	(setq sys-let-special 
	      (loop for x in sys-let-clauses
	         collect (if (atom x) x (car x))))
	(append! setq-vars sys-let-special)
	(append! '(backtracking-queue backtracking-state backtracking-flag
				      continuation-setq-flag)
		 sys-let-special)
	(setq sys-let-fixnum nil)
	(setq sys-declaration `(declare (special ,@ sys-let-special)))
	(append! `((,funcall-continue-var
		    #'(lambda ,user-vars
			,sys-declaration
			(continue ,@ user-vars))))
		 sys-let-clauses)
	)
      #-r7 (when sys-let-fixnum
	     (setq sys-declaration `(declare (fixnum ,@ sys-let-fixnum))))
      ;; Storage layout must be finalized before here.
      (when print-continuations-if
	(setq print-gensym-continuation
	  `(format t (format-template "~&    "
			(:separated-list "~A = ~S" ",~%    ")
			".~%")
	     (list ,@ (loop for v in spread-list
		         as g in setq-vars
			 when (member v user-vars)
		           collect `(quote ,v)
		           and collect g))))
	(setq print-gensym-continuation
	  `(let ((*print-circle* t))
             ,print-gensym-continuation))
	(setq print-spread-continuation
	  `(format t (format-template "~&    "
			(:separated-list "~A = ~S" ",~%    ")
			".~%")
	     (list ,@ (loop for v in spread-list
		         when (member v user-vars)
		           collect `(quote ,v)
		           and collect v))))
	(setq print-spread-continuation
	  `(let ((*print-circle* t))
	     ,print-spread-continuation))
	(setq print-setq-continuation
	  `(progn (format t "~&Continuing with~%")
		  ,print-gensym-continuation))
	(unless (eq print-continuations-if 't)
	   (setq print-setq-continuation
		 `(when ,print-continuations-if ,print-setq-continuation)))
	(setq print-enqueue-continuation
	  `(progn (format t "~&Enqueueing continuation with~%")
		  (let ,spread-list 
		    ,spread-form
		    ,print-spread-continuation)))
	(unless (eq print-continuations-if 't)
	   (setq print-enqueue-continuation
		 `(when ,print-continuations-if ,print-enqueue-continuation)))
	(setq print-pop-continuation
	      `(progn (format t "~&Continuing from queue with~%")
		      ,print-spread-continuation))
	(unless (eq print-continuations-if 't)
	  (setq print-pop-continuation
		`(when ,print-continuations-if ,print-pop-continuation)))
	(setq print-start-continuation
	      `(progn (format t "~&Starting with~%")
		      ,print-spread-continuation))
	(unless (eq print-continuations-if 't)
	  (setq print-start-continuation
		`(when ,print-continuations-if ,print-start-continuation)))
	)
      (setq continue-macro-body
	`(list 'cond
	   (list ',can-setq-test
		 '(setq continuation-setq-flag t)
		 ,setq-maker
		 ',print-setq-continuation
		 ',incf-setq-form)
	   (list 't
		 ,gather-maker
		 ',enqueue-form
		 ',print-enqueue-continuation
		 ',incf-enqueue-form)
	      ))
      (setq continue-macro `(continue ,user-vars ,continue-macro-body))
      (return
	`(macrolet (,continue-macro)
	   (catch 'continuation-search
	     (let ((backtracking-queue nil)
		   (backtracking-flag nil)
		   (continuation-setq-flag nil)
		   ,@ sys-let-clauses
		   ,@ setq-vars
		   backtracking-state)
	       ,sys-declaration
	       ,print-start-continuation
	       (loop 
		  do (setq continuation-setq-flag nil)
		     ,(progify `(,incf-step-form 
				 ,each-step-form
				 ,@ body-forms))
		     (cond (continuation-setq-flag
			    (setq backtracking-flag nil)
			    ,unsetq-form)
			   (t
			    (setq backtracking-flag t)
			    (unless (setq backtracking-state
					  (pop backtracking-queue))
			      ,finally-stats-form
			      ,@ finally-forms
			      (return nil))
			    ,spread-form
			    ,print-pop-continuation)
			   )))))))))

;;;
;;; Bitsets used in filtered letter trees.
;;;

(defun bitvectors-overlap-p (bv1 bv2)
  ;; true if these have a bit in common, i.e., if their bitvector-and
  ;; would have some nonzero element.  0-extend the smaller if they're of
  ;; different length, so the "extra" bits don't matter.
  (loop for index below (min (length bv1) (length bv2))
        for bv1-bit = (bit bv1 index)
        for bv2-bit = (bit bv2 index)
        thereis (and (not (zerop bv1-bit))
                     (not (zerop bv2-bit)))))

(defun bitvector-or (bv1 bv2)
  ;; take the boolean or of 2 bitvectors, 0-extending the smaller one.
  (loop with bv1-length = (length bv1)
        with bv2-length = (length bv2)
        with small-bv   = (if (< bv1-length bv2-length) bv1 bv2)
        with big-bv     = (if (< bv1-length bv2-length) bv2 bv1)
        with answer-bv = (make-array (max bv1 bv2) 
                                     :element-type 'bit
                                     :initial-contents big-bv)
        for i below (min bv1-length bv2-length)
        for bit = (bit small-bv i)
        when (not (zerop bit))
          do (setf (bit answer-bv i) 1)
        finally (return answer-bv)))
        
