;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

;;;
;;; A slightly odd mixin for pathnames.
;;;

(defclass filename-symbol-mixin (print-object-mixin)
  ((filename    :accessor filename
                :initform nil
                :initarg :filename)
   (defaultf    :accessor defaultf
                :initform nil
                :initarg :defaultf)
   (name-symbol :accessor name-symbol
                :initarg :name-symbol)))

;;; PCL::initialize???  Ick!  Someday, this should be initialize-instance.
(defmethod pcl::initialize :after ((self filename-symbol-mixin) init-plist)
  (declare (ignore init-plist))
  (when (filename self)
    ;; coerce the filename slot to a pathname
    (setf (filename self) (pathname (filename self))))
  (when (and (filename self)
             (defaultf self))
    ;; merge it with the default
    (setf (filename self) 
          (merge-pathnames (filename self) (pathname (defaultf self)))))
  (unless (or (and (slot-boundp self 'name-symbol)
                   (name-symbol self))
              (null (filename self)))
    ;; name-symbol isn't bound to something non-null and there is a filename
    (let* ((real-file (or (probe-file (filename self))
                          (filename self))))
      ;; construct a symbol from the filename or the real file.
      (setf (name-symbol self)
            (intern (format nil "~:@(~A-~A~)"
                            (pathname-name real-file)
                            (pathname-type real-file))))))
  (when (and (slot-boundp self 'name-symbol)
             (name-symbol self))
    ;; YUCK!  I CANNOT BELIEVE THIS! -- sgr
    (set (name-symbol self) self)))

(defmethod newest-filename ((self filename-symbol-mixin)
                            &aux (path (filename self)))
  ;; just like the existing filename, but with :newest version.
  (make-pathname :host      (pathname-host path)
                 :device    (pathname-device path)
                 :directory (pathname-directory path)
                 :name      (pathname-name path)
                 :type      (pathname-type path)
                 :version   :newest))

;;; These should be used with in conjunction with WITH-OPEN-STREAM.

(defmethod input-stream ((self filename-symbol-mixin))
  ;; input stream corresponding to the newest version of the file
  (let ((stream (open (newest-filename self) :direction :input)))
    (setf (filename self)  (truename stream))
    stream))

(defmethod output-stream ((self filename-symbol-mixin))
  ;; output stream corresponding to the newest version of the file
  (let ((stream (open (newest-filename self) :direction :output)))
    (setf (filename self) (truename stream))
    stream))

(defmethod file-syntax-error ((self filename-symbol-mixin) fmt &rest args)
  ;; how to register a complaint
  (error "Syntax error in file ~A:~&~?" (filename self) fmt args))

(defmethod print-object-name ((self filename-symbol-mixin))
  ;; supply the idiosyncratic part of our printer.
  (if (slot-boundp self 'name-symbol)
    (name-symbol self)
    "-- No file --"))
