;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

;;;
;;; This flavor implements the actual generation machine.
;;;

(defclass kimmo-generator (kimmo-machine) ())

(defmethod lex-lookup ((self kimmo-generator) lex)
  ;; method not used in actual operation.
  ;; returns an alist associating surf chars with vector lists.
  (cdr (assoc lex (machine-data self))))

(defmethod absorb-machine-data ((self kimmo-generator) feasible-pair-entries)
  (setf (machine-data self) nil)
  (loop for (pair . vector-list) in feasible-pair-entries
        as (lex . surf) = pair
        do (set-machine-vectors self lex surf vector-list)))

(defmethod generate ((self kimmo-generator) char-list tracing)
  (check-type char-list list)
  (setf (results self) nil)
  (if tracing
      (generate-internal-tracing self char-list (get-tracer self tracing))
      (generate-internal self char-list))
  (results self))

(defmethod generate-internal ((self kimmo-generator) char-list)
  (let ((revresult nil)
	(config (initial-config self))
	(remainder char-list))
    (continuation-search (revresult config remainder)
      (cond ((null remainder)
             (when (config-final self config)
               (push (make-surface-string 
                      (automaton-set self) (reverse revresult))
                     (results self))))
            (t 
             (loop for (surf . vector-list)
                   in (cdr (assoc (car remainder) (machine-data self)))
                   as advanced = (advance-config self config vector-list)
                   when advanced
                     do (continue (cons surf revresult) advanced 
                                  (cdr remainder))))))))

(defmethod generate-internal-tracing ((self kimmo-generator) char-list tracer)
  (let ((revresult nil)
	(config (initial-config self))
	(remainder char-list)
	(source `(:start-generation ,char-list))
	choice-point
	result
	finalities)
    (trace-init tracer)
    (macrolet ((trace-step* (step &optional (result-chars '(reverse revresult)))
                 ;; *** isn't it a bug to have these named the same??? -- sgr
		 `(trace-step tracer ,step
                              continuation-level choice-point nil
                              ,result-chars))
	       (trace-source-step (step)
		 `(trace-step tracer ,step
                              (1- continuation-level)
                              choice-point backtracking-flag
                              (reverse revresult)))
	       (silent ()
		 '(and (not continuation-setq-flag)	;; works for depth first
		       (= choice-point (step-number tracer)))))
      (continuation-search (revresult config remainder source choice-point)
	(:include-level)
	(:print-statistics-if *kimmo-machine-print-statistics-flag*)
	;; (:print-continuations-if t)
	;; (:print-steps)
	(trace-source-step source)
	(setq choice-point (step-number tracer))
	(cond ((not (null remainder))
	       (loop for (surf . vector-list)
		  in (cdr (assoc (car remainder) (machine-data self)))
		     as advanced = (advance-config-0 self config vector-list)
		     if (not (member 0 advanced))
		       do (continue (cons surf revresult)
				    advanced (cdr remainder)
				    `(:continue-advance 
				          ,(cons (car remainder) surf)
					  ,advanced)
				    choice-point)
		     else 
		       do (trace-step* `(:reject-automata 
					,(cons (car remainder) surf)
					,config ,advanced
					,(automaton-names self))
				      (reverse (cons surf revresult)))))
	      ((member nil 
                       (setq finalities
                             (config-finalities self config)))
	       (trace-step* `(:reject-nonfinal-config nil ,config
			     ,finalities ,(automaton-names self))))
	      (t (setq result (make-surface-string (automaton-set self)
                                                   (reverse revresult)))
		 (push result (results self))
		 (trace-step* `(:accept-result nil ,result))))
	(when (silent)
	  (trace-step* '(:nothing-to-do)))
	))
    (trace-dump tracer)
    ))

