;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

;;;
;;; A data structure on which lexicons are built.
;;;

(defstruct (letter-tree-node
            (:type list))
  ;; Generally for internal use only; send messages to the letter tree. However,
  ;; code may depend on the fact that each node is a list with a key in its CAR.
  key 
  (outputs nil)
  (subtrees nil)
  (plist nil))

(defclass letter-tree ()
  ((root :accessor root)
   (null-char :initform '\0
              :accessor null-char)))

(defmethod node-outputs ((self letter-tree) node &optional filter-bits)
  (declare (ignore filter-bits))
  (letter-tree-node-outputs node))

(defmethod all-node-outputs ((self letter-tree) node)
  (letter-tree-node-outputs node))

(defmethod node-branches ((self letter-tree) node)
  (letter-tree-node-subtrees node))

(defmethod get-branch ((self letter-tree) char node &optional filter-bits)
  (declare (ignore filter-bits))
  (assoc char (letter-tree-node-subtrees node)))

(defmethod node-plist ((self letter-tree) node)
  (let ((plist (letter-tree-node-plist node)))
    (when (null plist)
      ;; headed list so we can use getf on its cdr
      (setq plist (list nil))
      (setf (letter-tree-node-plist node) plist))
    plist))

(defmethod get-node-property ((self letter-tree) node prop)
  (let ((plist (letter-tree-node-plist node)))
    (when plist
      (getf (cdr plist) prop))))

(defmethod set-node-property ((self letter-tree) node prop val)
  (let ((plist (letter-tree-node-plist node)))
    (when (null plist)
      (setq plist (list nil))
      (setf (letter-tree-node-plist node) plist))
    (setf (getf (cdr plist) prop) val)))

(defmethod pcl::initialize :after ((self letter-tree) plist)
  (declare (ignore plist))
  (reset self))

(defmethod make-node ((self letter-tree) key)
  (make-letter-tree-node :key key))

(defmethod reset ((self letter-tree))
  (setf (root self) (make-node self 'root)))

(defmethod add-output ((self letter-tree) char-list output &optional filter-bits)
  ;; assumes the output list is really a list
  (declare (ignore filter-bits))
  (let* ((subtree (add-char-list self char-list))
	 (outlist (letter-tree-node-outputs subtree)))
    (unless (member output outlist :test #'equal)
      ;; this would be pushnew, except it goes on the end -- sgr.
      (setf (letter-tree-node-outputs subtree)
	    (nconc outlist (list output))))
    output))

(defmethod set-output-list ((self letter-tree) char-list output-list)
  (let* ((subtree (add-char-list self char-list)))
    (setf (letter-tree-node-outputs subtree) output-list)
    output-list))

(defmethod look-up-char-list ((self letter-tree) char-list
                              &optional root-subtree filter-bits)
  (declare (ignore filter-bits))
  (loop with tree = (or root-subtree (root self))
        for x in char-list
        as subtrees = (letter-tree-node-subtrees tree)
        as subtree = (assoc x subtrees)
        do (cond ((eql x (null-char self))
                  ;; not a true character; ignore
                  nil)
                 ((not (null subtree))
                  (setq tree subtree))
                 (t (return nil)))
        finally (return tree)))

(defmethod splice-in-subtree ((self letter-tree) tree subtree)
  ;; there may be daemons on this method
  (setf (letter-tree-node-subtrees tree)
	(nconc (letter-tree-node-subtrees tree)
	       (list subtree))))

(defmethod add-char-list ((self letter-tree) char-list 
                          &optional root-subtree filter-bits)
  (declare (ignore filter-bits))
  (loop with tree = (or root-subtree (root self))
        for x in char-list
        for subtrees = (letter-tree-node-subtrees tree)
        for subtree = (assoc x subtrees)
        do (cond ((eql x (null-char self))
                  ;; not a true character; ignore
                  nil)
                 ((null subtree)
                  (setq subtree (make-node self x))
                  (splice-in-subtree self tree subtree)
                  (setq tree subtree))
                 (t (setq tree subtree)))
        finally (return tree)))

(defvar *letter-tree-mapping-extf*)
(defvar *letter-tree-mapping-targf*)
(defvar *letter-tree-mapping-nullp*)
(defvar *letter-tree-mapping-selfp*)
(defvar *letter-tree-mapping-subf*)
(defvar *letter-tree-mapping-sortf*)

(defmethod %apply-tree-mapping-function ((self letter-tree) node rclist)
  (let ((target (if (null *letter-tree-mapping-targf*)
		    node
		    (funcall *letter-tree-mapping-targf* node))))
    (when (or target *letter-tree-mapping-nullp*)
      (if (eq rclist 'none)
	  (if *letter-tree-mapping-selfp*
	      (funcall *letter-tree-mapping-extf* target self)
	      (funcall *letter-tree-mapping-extf* target))
	  (if *letter-tree-mapping-selfp*
	      (funcall *letter-tree-mapping-extf* (reverse rclist) target self)
	      (funcall *letter-tree-mapping-extf* (reverse rclist) target))))))

(defmethod map-over-subtrees ((self letter-tree) extended-function 
                              &key
                              ((:root root-subtree) (root self))
                              supply-char-lists 
                              supply-letter-tree
                              (subtrees-function #'(lambda (x)
                                                     (node-branches self x)))
                              &allow-other-keys)
  ;; args of fcn are ([char-list] node [letter-tree]).
  (let ((*letter-tree-mapping-extf* extended-function)
        (*letter-tree-mapping-targf* nil)
        (*letter-tree-mapping-nullp* t)
        (*letter-tree-mapping-selfp* supply-letter-tree)
        (*letter-tree-mapping-subf* subtrees-function))
    (%map-over-subtrees-internal self root-subtree 
                                 (if supply-char-lists nil 'none))))

(defmethod %map-over-subtrees-internal ((self letter-tree) subtree rclist)
  (%apply-tree-mapping-function self subtree rclist)
  (dolist (st (funcall *letter-tree-mapping-subf* subtree))
    (%map-over-subtrees-internal self st 
                                 (if (eq rclist 'none)
                                   rclist
                                   (cons (car st) rclist)))))

(defmethod sort-all-subtrees ((self letter-tree) predicate)
  (let ((*letter-tree-mapping-sortf* predicate))
    (map-over-subtrees self
                       #'(lambda (node)
                           (setf (letter-tree-node-subtrees node)
                                 (sort (letter-tree-node-subtrees node)
                                       *letter-tree-mapping-sortf*
                                       :key #'car))))))

(defmethod map-output-lists ((self letter-tree) extended-function
                             &key
                             ((:root root-subtree) (root self))
                             (outputs-function #'(lambda (x) 
                                                   (all-node-outputs self x)))
                             supply-char-lists
                             supply-letter-tree
                             include-null-output-lists
                             (subtrees-function #'(lambda (x)
                                                    (node-branches self x))))
  ;; args of fcn are ([char-list] outputs [letter-tree]).
  (let ((*letter-tree-mapping-extf* extended-function)
        (*letter-tree-mapping-targf* outputs-function)
        (*letter-tree-mapping-nullp* include-null-output-lists)
        (*letter-tree-mapping-selfp* supply-letter-tree)
        (*letter-tree-mapping-subf* subtrees-function))
    (%map-over-subtrees-internal self root-subtree 
                                 (if supply-char-lists nil 'none))))

(defmethod print-outputs ((self letter-tree) &rest keywords)
  ;; assumes the output list is a list
  (apply #'map-output-lists 
         self
         #'(lambda (char-list outputs)
             ;; Wow; TECO lives!  "~&~{~A~}:~20T~{~S~#[~:;, ~]~}~%"
             (format t (format-template "~&" 
                                        (:use-repeatedly-on-list "~A")
                                        ":~20T"
                                        (:separated-list "~S" ", ")
                                        "~%")
                     char-list
                     outputs))

     :supply-char-lists t
     keywords))
