;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

;;;
;;; Lexicons for Kimmo.  Many lexicons go together to make a dictionary.
;;;

(defclass kimmo-lexicon-base-mixin ;base-mixin?? -- sgr
  (print-object-mixin) ;; not KIMMO-STRING-MIXIN any more
  ((name :initarg :name
         :initform "Unnamed"
         :accessor name)
   ;; backpointer to kimmo-dictionary
   (dictionary :initarg :dictionary
               :accessor dictionary)))

(defmethod map-stored-entries ((self kimmo-lexicon-base-mixin) extended-function)
  ;; a default method.  Extended-function takes one argument, OUTPUTS.
  (map-output-lists self extended-function))

(defmethod show-entries ((self kimmo-lexicon-base-mixin))
  ;; default method.  An output is initially a list of the form:
  ;; (<continuation-class-name> <features> <string>)
  ;; but the class-name is later replaced by a list (class-name . lexicon-roots)
  (format t "~&Lexicon ~A:~
             ~&String~15TCClass~25TFeatures~
             ~&======~15T======~25T========"
          (name self))
  (map-stored-entries 
   self
   #'(lambda (outputs)
       (loop for (cclass feats string) in outputs
             do (format t "~&~A  ~15T~A  ~25T~:A~%" 
                        string
                        (if (listp cclass)	; already expanded
                          (car cclass)
                          cclass)
                        (reverse feats)))))
  (terpri t) ; so show-lexicon-entries on a dictionary will separate lexicons
  (values))

(defmethod process-file-entry ((self kimmo-lexicon-base-mixin) stream)
  ;; file input
  (disable-digits *kimmo-reader*)
  (mapc #'(lambda (char) (enable-alphabetic-char *kimmo-reader* char))
        '( #\: #\;)) ;#\.
  ;; the next 3 reads cause an error if they hit an EOF.
  (let* ((entry        (read-next-thing *kimmo-reader* stream nil))
         (suffix-class (read-next-thing *kimmo-reader* stream nil))
         (feat-string  (read-next-thing *kimmo-reader* stream nil)))
    (unless (symbolp entry)
      (file-syntax-error self "~S is an invalid entry." entry))
    (unless (class-defined (dictionary self) suffix-class)
      (file-syntax-error self "~S is an invalid suffix class." suffix-class))
    (let ((feat-list (etypecase feat-string
                       (list feat-string)
                       (string (reverse (listify-feature-string 
                                         (dictionary  self) feat-string))))))
      (add-output self (explode-string (dictionary self) entry)
                  (list suffix-class feat-list entry)))))

;;;
;;; Miscellaneous
;;;

(defmethod print-object-name ((self kimmo-lexicon-base-mixin))
  ;; the idiosyncratic part of printing one of these
  (name self))

(defmethod file-syntax-error ((self kimmo-lexicon-base-mixin) fmt &rest args)
  (error "Syntax error while reading lexicon ~A:~&~?" (name self) fmt args))

;;;
;;; Instantiable lexicons.
;;;

(defclass kimmo-lexicon (letter-tree kimmo-lexicon-base-mixin) ())

(defmethod reset :after ((self kimmo-lexicon))
  (setf (car (root self)) self))

;; the standard lexicon doesn't do any filtering because the lexicons
;; are already split up
(defmethod possible-branches ((self kimmo-lexicon) node)
  (node-branches self node))
(defmethod filtered-branches ((self kimmo-lexicon) node)
  (node-branches self node))

(defmethod filter-branch ((self kimmo-lexicon) branch)
  branch)

(defmethod zap-output-list-continuation-classes ((self kimmo-lexicon) L)
  ;; see the note in the show-entries method, above.  We make the replacement:
  ;; in (class-name feats string) -> ((class-name . lex-roots) feats string)
  (dolist (out L)
    ;; each element is an "output"
    (unless (listp (car out))	; already expanded
      (setf (car out) 
	    (cons (car out)
                  (retrieve-class-lexicon-roots (dictionary self) 
                                                (car out)))))))

(defmethod zap-continuation-classes ((self kimmo-lexicon))
  (map-stored-entries 
   self 
   #'(lambda (l)
       (zap-output-list-continuation-classes self l))))
