;;; -*- Mode: Lisp; Package: Kimmo; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package 'kimmo)

;;;
;;; This flavor is the base for the recognizer and generator machines.
;;;

(defclass kimmo-machine (tracer-storage-mixin)
  ;; backpointer to kimmo-automaton-set
  ((automaton-set :accessor automaton-set
                  :initarg automaton-set
                  :initform nil)
   ;; a list of automata
   (automaton-list :accessor automaton-list
                   :initarg automaton-list
                   :initform nil)
   (automaton-names :accessor automaton-names
                    :initform 'computed)
   ;; a list of T-type arrays
   (finality-vectors :accessor finality-vectors
                     :initarg :finality-vectors
                     :initform nil)
   ;; a list of states
   (initial-config :accessor initial-config
                   :initarg :initial-config
                   :initform nil)
   ;; a two-level alist, with primary lookup by either SurfChar or LexChar,
   ;; depending on machine type; final value is a list of transition vectors.
   (machine-data :accessor machine-data
                 :initarg :machine-data
                 :initform nil)
   ;; accumulates results during processing
   (results :accessor results
            :initarg :results
            :initform nil)
   (default-tracer-flavor :accessor default-tracer-flavor
     :initarg :defalt-tracer-flavor
     :initform 'kimmo-tracer)
     ))

(defvar *kimmo-machine-print-statistics-flag* t)

(defmethod make-tracer ((self kimmo-machine) flavor-name)
  (make-instance flavor-name :machine self))

(defmethod pcl::initialize :after ((self kimmo-machine) plist)
  (declare (ignore plist))
  (or (automaton-set self)
      (error "Automaton-set is NIL in ~S." self))
  (or (initial-config self)
      (error "Initial-config is NIL in ~S." self))
  (or (finality-vectors self)
      (error "Finality-vectors is NIL in ~S." self))
  (unless (eq (automaton-names self) 'computed)
    (error "The AUTOMATON-NAMES slot of ~S has incorrectly been initialized."
           self))
  (setf (automaton-names self) 
	(mapcar #'name (or (automaton-list self)
                           (error "Automaton-list is NIL in ~S." self)))))

(defmethod print-object-name ((self kimmo-machine))
  ;; idiosyncratic part of printing one of these
  (name-symbol (automaton-set self)))

(defmethod get-machine-data-cell ((self kimmo-machine) key1 key2)
  (let ((cell1 (assoc key1 (machine-data self))))
    (when (null cell1)
      (setq cell1 (list key1))
      (setf (machine-data self) (nconc (machine-data self) (list cell1))))
    (let ((cell2 (assoc key2 (cdr cell1))))
      (when (null cell2)
	(setq cell2 (list key2))
	(setf (cdr cell1) (nconc (cdr cell1) (list cell2))))
      cell2)))

(defmethod set-machine-vectors ((self kimmo-machine) key1 key2 vectors)
  (unless (null vectors)
    (let ((cell (get-machine-data-cell self key1 key2)))
      (when (cdr cell)
	(error "Duplicate machine-data entry for ~S + ~S." key1 key2))
      (setf (cdr cell) vectors))))

(defmethod advance-config-0 ((self kimmo-machine) state-list vector-list)
  (loop for i in state-list
        as v in vector-list
        as j = (aref v i)
        collect j))

(defmethod advance-config ((self kimmo-machine) state-list vector-list)
  (loop for i in state-list
        as v in vector-list
        as j = (aref v i)
        when (= j 0)
          return nil
        collect j))

(defmethod config-finalities ((self kimmo-machine) state-list)
  (loop for i in state-list
        as finality-vector in (finality-vectors self)
        collect (aref finality-vector i)))

(defmethod config-final ((self kimmo-machine) state-list)
  (loop for i in state-list
        as finality-vector in (finality-vectors self)
        always (aref finality-vector i)))

(defmethod config-nonfinal-automaton ((self kimmo-machine) state-list)
  (loop for i in state-list
        as aut in (automaton-list self)
        as finality-vector in (finality-vectors self)
        unless (aref finality-vector i)
          return aut
        finally (error "Config ~D is final." state-list)))

(defmethod format-finality-complaint ((self kimmo-machine) dst config)
  (loop for i in config
        as finality-vector in (finality-vectors self)
        as name in (automaton-names self)
        when (not (aref finality-vector i))
          collect (list i name) into complainants
        finally
          (cond ((null complainants)
                 (error "Config ~D is final." config))
                ((null (cdr complainants))
                 (return
                  (format dst "state ~D of ~A is non-final" 
                          (caar complainants) (cadar complainants))))
                (t (return
                    (format dst
                            (format-template "states "
                                             (:separated-list 
                                              (:use-repeatedly-on-list "~D of ~A") ", ")
                                             "are non-final")
                            complainants))))))

(defmethod config-blockage-automaton ((self kimmo-machine) state-list)
  (loop for j in state-list
        as aut in (automaton-list self)
        when (= j 0)
          return aut
        finally (error "Config ~D is not blocked." state-list)))

(defmethod format-blockage-complaint ((self kimmo-machine) dst config nconfig)
  (loop for i in config
        as j in nconfig
        as name in (automaton-names self)
        when (= j 0) 
          collect (list name i) into complainants
        finally
          (cond ((null complainants)
                 (error "New config ~D is not blocked." nconfig))
                ((null (cdr complainants))
                 (return 
                  (format dst "automaton ~A blocks from state ~D"
                          (caar complainants) (cadar complainants))))
                (t (return
                    (format dst
                            (format-template "automata block: "
                                             (:separated-list
                                              (:use-repeatedly-on-list "~A from state ~D") ", ")
                                             )
                            complainants))))))

;;;
;;; Trace stuff.
;;;

(defclass kimmo-tracer ()
  ((step-number :accessor step-number
                :initform 0)
   (machine :accessor machine
            :initarg :machine)))

(defmethod trace-init :before ((self kimmo-tracer))
  (setf (step-number self) 0))

(defmethod trace-init ((self kimmo-tracer))
  ;; *** I can't actually find this in the source
  nil)

(defmethod trace-step :after ((self kimmo-tracer)
                              step level choice-point
                              backtrack-flag result-chars)
  (declare (ignore step level choice-point backtrack-flag result-chars))
  (incf (step-number self)))

(defmethod trace-dump ((self kimmo-tracer))
  ;; default method
  nil)

(defmethod trace-step ((self kimmo-tracer)
                       step level choice-point
                       backtrack-flag result-chars)
  ;; default method
  (declare (ignore result-chars))
  (let (char-list entry config finalities nconfig aut-names result lexicon pair)
    (unless (member (car step) '(:start-recognition :start-generation))
      (format *trace-output* "~D.~5T(~D)~10T" (step-number self) level)
      (cond (backtrack-flag
	      (format *trace-output* "[~D]" choice-point))
	    ((not (= (step-number self) choice-point))
	     (format *trace-output* " + ")))
      (format *trace-output* "~vT" (+ (* 2 level) 17.)))
    (ecase (car step)
      (:start-recognition
       (setq char-list (second step))
       (format *trace-output*
               (format-template "~&Recognizing surface form "
                                "\"" (:use-repeatedly-on-list "~A") "\"" ".~%")
               char-list))
      (:start-generation
       (setq char-list (second step))
       (format *trace-output*
               (format-template "~&Generating from lexical form "
                                "\"" (:use-repeatedly-on-list "~A") "\"" ".~%")
               char-list))
      (:reject-leftover-input
       (setq entry (second step)
             char-list (third step))
       (format *trace-output*
               "Entry ~S is word-final --> path rejected (leftover input).~%"
               entry))
      (:reject-nonfinal-config
       (setq entry (second step)
             config (third step)
             finalities (fourth step)
             aut-names (fifth step))
       (format *trace-output*
               "End --> rejection because ~A.~%"
               (format-finality-complaint (machine self) nil config)))
      (:accept-result
       (setq entry (second step)
             result (third step))
       (format *trace-output*
               "End --> result is ~S.~%" result))
      (:reject-automata
       (setq pair (second step)
             config (third step)
             nconfig (fourth step)
             aut-names (fifth step))
       (format *trace-output*
               "(~A.~A) --> ~A.~%"
               (car pair) (cdr pair)
               (format-blockage-complaint (machine self) nil config nconfig)))
      (:continue-new-lexicon
       (setq entry (second step)
             lexicon (third step)
             config (fourth step))
       (format *trace-output*
               "Entry ~S ends --> new lexicon ~A, config ~D~%"
               entry (name lexicon) config))
      (:continue-advance
       (setq pair (second step)
             config (third step))
       (format *trace-output* "(~A.~A) --> ~D.~%" (car pair) (cdr pair) config))
      (:nothing-to-do
       (format *trace-output* "Nothing to do.~%")))))


