;;; -*- Mode: Lisp; Package: Kimmo; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package 'kimmo)

;;;
;;; This implements the actual recognizer machine.
;;;

(defclass kimmo-recognizer (kimmo-machine) ())

(defmethod surf-lookup ((self kimmo-recognizer) surf)
  ;; method not used in actual operation.
  ;; returns an alist associating lex chars with vector lists.
  (cdr (assoc surf (machine-data self))))

(defmethod absorb-machine-data ((self kimmo-recognizer) feasible-pair-entries)
  (setf (machine-data self) nil)
  (loop for (pair . vector-list) in feasible-pair-entries
     as (lex . surf) = pair
     do (set-machine-vectors self surf lex vector-list)))

(defmethod recognize ((self kimmo-recognizer) char-list dic tracing)
  (check-type char-list list)
  ;; flush the results of any previous run
  (setf (results self) nil)
  (let ((root (root-lexicon-root dic))
        (tracer (when tracing (get-tracer self tracing))))
    (if tracing
      (recognize-internal-tracing self char-list root tracer)
      (recognize-internal         self char-list root)))
  (results self))

(defmethod recognize-internal ((self kimmo-recognizer) char-list root)
  ;; letter tree output is ((cont-class-name . new-roots) features)
  (let* ((revresult nil)
         (config (initial-config self))
         (remainder char-list)
         (node root)
         (lexicon (car root))
         (revfeats nil)
         outputs)
    (continuation-search (revresult config remainder lexicon node revfeats)
      (setq outputs (node-outputs lexicon node))
      ;; "If lexicon starts with an entry...."
      (loop with result
            for ((class-name . new-roots) features) in outputs
            if (null new-roots)	; no continuation classes
            do (when (and (null remainder)
                          (config-final self config))
                 (setq result (make-lexical-string (automaton-set self)
                                                   (reverse revresult)))
                 (setq result 
                       (list result
                             (reverse (append features revfeats))))
                 (push result (results self)))
            else do (loop for new-root in new-roots
                          do (continue
                              revresult config remainder 
                              (car new-root) new-root
                              (append features revfeats))))
      (loop with (surf lex-level vector-list advanced)
            for branch in (possible-branches lexicon node)
            as lex = (car branch)
            when (filter-branch lexicon branch)
            ;; Lambda-processing first.
            do (when (and (setq surf '\0)
                          (setq lex-level (assoc surf (machine-data self)))
                          (setq vector-list (cdr (assoc lex (cdr lex-level))))
                          (setq advanced (advance-config self 
                                                         config vector-list)))
                 (continue (cons lex revresult)
                           advanced remainder lexicon branch revfeats))
            ;; Normal processing next.
            (when (and (setq surf (car remainder))
                       (setq lex-level (assoc surf (machine-data self)))
                       (setq vector-list (cdr (assoc lex (cdr lex-level))))
                       (setq advanced (advance-config self 
                                                      config vector-list)))
              (continue (cons lex revresult)
                        advanced (cdr remainder) lexicon branch revfeats))))))

(defmethod recognize-internal-tracing ((self kimmo-recognizer)
                                       char-list root tracer)
  ;; letter tree output is ((cont-class-name . new-roots) features)
  (let* ((revresult nil)
         (config (initial-config self))
         (remainder char-list)
         (node root)
         (lexicon (car root))
         (revfeats nil)
         (source `(:start-recognition ,char-list))
         choice-point
         outputs
         finalities)
    (trace-init tracer)
    (macrolet ((trace-step* (step &optional (result-chars '(reverse revresult)))
                 `(trace-step tracer ,step 
                              continuation-level choice-point nil ,result-chars))
               (trace-source-step (step)
                 `(trace-step tracer ,step
                              (1- continuation-level)
                              choice-point backtracking-flag
                              (reverse revresult)))
               (silent ()
                 '(and (not continuation-setq-flag)	;; works for depth first
                       (= choice-point (step-number tracer)))))
      (continuation-search (revresult config remainder lexicon node revfeats
                                      source choice-point)
        (:include-level)
        (:print-statistics-if *kimmo-machine-print-statistics-flag*)
        ;; (:print-continuations-if t)
        ;; (:print-steps)
        (trace-source-step source)
        (setq choice-point (step-number tracer))
        (setq outputs (node-outputs lexicon node))
        ;; "If lexicon starts with an entry...."
        (loop with result
              for ((class-name . new-roots) features entry) in outputs
              if (null new-roots)	
              ;; There are no continuation-classes (#), so either this 
              ;; is terminal, or we fail & backtrack.
              do (cond ((not (null remainder))
                        (trace-step* `(:reject-leftover-input ,entry ,remainder)))
                       ((member nil 
                                (setq finalities 
                                      (config-finalities self config)))
                        (trace-step* `(:reject-nonfinal-config ,entry ,config
                                                               ,finalities ,(automaton-names self))))
                       (t (setq result 
                                (make-lexical-string
                                 (automaton-set self) (reverse revresult)))
                          (setq result 
                                (list result
                                      (reverse (append features revfeats))))
                          (trace-step* `(:accept-result ,entry ,result))
                          (push result (results self))))
              else do (loop for new-root in new-roots
                            do (continue
                                revresult config remainder 
                                (car new-root) new-root
                                (append features revfeats)
                                `(:continue-new-lexicon ,entry ,(car new-root)
                                                        ,config)
                                choice-point)))
        (loop with (surf lex-level vector-list advanced)
              for branch in (possible-branches lexicon node)
              as lex = (car branch)
              when (filter-branch lexicon branch)
              do ;; Lambda-processing first.
              (setq surf '\0)
              (cond ((null (setq lex-level (assoc surf (machine-data self))))
                     ;; machine permits no null elements here
                     nil)
                    ((null (setq vector-list (cdr (assoc lex (cdr lex-level)))))
                     ;; machine doesn't permit (lex.0) here
                     nil)
                    ((not 
                      (member 0 
                              (setq advanced
                                    (advance-config-0 self config vector-list))))
                     (continue (cons lex revresult)
                               advanced remainder lexicon branch revfeats
                               `(:continue-advance ,(cons lex surf) 
                                                   ,advanced)
                               choice-point))
                    (t
                     (trace-step* `(:reject-automata ,(cons lex surf)
                                                     ,config ,advanced
                                                     ,(automaton-names self))
                                  (reverse (cons lex revresult)))))
              ;; Normal processing next.
              (cond ((null (setq surf (car remainder)))
                     nil)
                    ((null (setq lex-level (assoc surf (machine-data self))))
                     ;; machine forbids current surface element here
                     nil)
                    ((null 
                      (setq vector-list (cdr (assoc lex (cdr lex-level)))))
                     ;; machine forbids (lex.surf) pair here
                     nil)
                    ((not 
                      (member 0 
                              (setq advanced
                                    (advance-config-0 self config vector-list))))
                     (continue (cons lex revresult)
                               advanced (cdr remainder) lexicon
                               branch revfeats
                               `(:continue-advance ,(cons lex surf) 
                                                   ,advanced)
                               choice-point))
                    (t 
                     (trace-step* `(:reject-automata ,(cons lex surf)
                                                     ,config ,advanced
                                                     ,(automaton-names self))
                                  (reverse (cons lex revresult))))))
        (when (silent)
          (trace-step* '(:nothing-to-do)))
        ))
    (trace-dump tracer)
    ))
