;;; -*- Mode: Lisp; Package: KIMMO; Syntax: Common-Lisp -*-
;;; Converted to CL + PCL January 1989 by sgr from a ZetaLisp version.
;;;

;; Copyright (C) 1990 Robert C. Berwick

;; This file is part of the 6.863 Software

;; The 6.863 Software is distributed in the hope that it will be
;; useful, but WITHOUT ANY WARRANTY.  No author or distributor
;; accepts responsibility to anyone for the consequences of using it
;; or for whether it serves any particular purpose or works at all,
;; unless he says so in writing.  Refer to the 6.863 Software General
;; Public License for full details.

;; Everyone is granted permission to copy, modify and redistribute the
;; 6.863 Software, but only under the conditions described in the 6.863
;; Software General Public License.   A copy of this license is
;; supposed to have been given to you along with the 6.863 Software so you
;; can know your rights and responsibilities.  It should be in a
;; file named COPYING.  Among other things, the copyright notice
;; and this notice must be preserved on all copies.

(in-package "KIMMO")

;;;
;;; This file implements some peculiar string operations
;;; such that I'm not sure where they belong.
;;;

(defclass kimmo-string-mixin ()
  ((strings-lower-case :initform nil
                       :accessor strings-lower-case)
   ;; sequence of trailing accents that can be incorporated into character
   (accent-codes       :initform nil
                       :accessor accent-codes)))

;;; PCL::INITIALIZE??  Will one day be INITIALIZE-INSTANCE, I hope.
(defmethod pcl::initialize :after ((self kimmo-string-mixin) plist)
  (declare (ignore plist))
  (get-defined-value '*kimmo-reader*))

(defmethod decompose-code-list ((self kimmo-string-mixin) codes)
  ;; This now allows for accents as well as for multi-char sequences
  ;; delimited by '<' and '>'.  It returns two values similar to the
  ;; "CAR" and "CDR" of the code list.
  (declare (values first tail))
  ;; actually, we want to say something like (every #'characterp codes)?
  (check-type codes (not null))
  (let ((first (cons (car codes) nil))
	(tail (cdr codes)))
    (when (strings-lower-case self)
      (setf (car first) (char-upcase (car first))))
    (when (char-equal (car first) #\<)
      ;; this marks a sequence of characters.  Collect them as though one char.
      (loop while tail
            for y = (pop tail)
            do (nconc first (cons y nil))
            until (char-equal y #\>)))
    ;; now collect the trailing accents after the char.
    (loop while (and tail (member (car tail) (accent-codes self)))
       do (nconc first (cons (pop tail) nil)))
    ;; return the (possibly composite) char with some accents, and the rest
    ;; of the characters from which to get more.
    (values first tail)))

(defmethod explode-string ((self kimmo-string-mixin) string)
  ;; this allows for multi-char sequences delimited by '<' and '>'
  (check-type string (or symbol string))
  (setq string (string string)) ;!
  (loop with first = nil
	with rest = (coerce string 'list) ;(exploden string) wants char-codes?
        while rest
        doing (multiple-value-setq (first rest) (decompose-code-list self rest))
        ;; do we *REALLY* want to implode these things into symbols, not strings?
        collect (intern (coerce first 'string))))

(defun kimmo-reader-macro-< (&optional (stream *standard-input*))
  ;; this is the corresponding reader macro.  DON'T USE IT!
  ;; IT TERMINATES ATOMS SO THAT ho<s,>+DIr READS AS TWO.
  (loop for x = (tyi stream)
        collecting x into answer
        until (char-equal x #\>)
        finally (return (coerce (cons #\< answer) 'string))))

(defmethod listify-feature-string ((self kimmo-string-mixin) string)
  (with-input-from-string (s (string-append "(" string ")"))
    (read-next-thing *kimmo-reader* s)))

(defmethod concat-all ((self kimmo-string-mixin) strings)
  ;; ugly as it may be, the "strings" are actually symbols!
  (apply #'concatenate 'string (mapcar #'string strings)))

(defmethod make-surface-string ((self kimmo-string-mixin) chars)
  (concat-all self (loop for x in chars
                         unless (eq x '\0) ; gack! (char-equal x #\Null)
                         collect x)))

(defmethod make-lexical-string ((self kimmo-string-mixin) chars)
  (concat-all self chars))

